"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = __importDefault(require("lodash"));
const child_process_1 = require("child_process");
const path_1 = __importDefault(require("path"));
const status_1 = __importDefault(require("../../status"));
const helpers_1 = require("../helpers");
const messages_1 = require("@cucumber/messages");
const value_checker_1 = require("../../value_checker");
const stopwatch_1 = require("../stopwatch");
const runWorkerPath = path_1.default.resolve(__dirname, 'run_worker.js');
class Coordinator {
    constructor({ cwd, eventBroadcaster, eventDataCollector, pickleIds, options, supportCodeLibrary, supportCodePaths, supportCodeRequiredModules, }) {
        this.cwd = cwd;
        this.eventBroadcaster = eventBroadcaster;
        this.eventDataCollector = eventDataCollector;
        this.stopwatch = options.predictableIds
            ? new stopwatch_1.PredictableTestRunStopwatch()
            : new stopwatch_1.RealTestRunStopwatch();
        this.options = options;
        this.supportCodeLibrary = supportCodeLibrary;
        this.supportCodePaths = supportCodePaths;
        this.supportCodeRequiredModules = supportCodeRequiredModules;
        this.pickleIds = pickleIds;
        this.nextPickleIdIndex = 0;
        this.success = true;
        this.workers = {};
        this.supportCodeIdMap = {};
    }
    parseWorkerMessage(worker, message) {
        if (value_checker_1.doesHaveValue(message.supportCodeIds)) {
            this.saveDefinitionIdMapping(message.supportCodeIds);
        }
        else if (message.ready) {
            this.giveWork(worker);
        }
        else if (value_checker_1.doesHaveValue(message.jsonEnvelope)) {
            const envelope = messages_1.messages.Envelope.fromObject(JSON.parse(message.jsonEnvelope));
            this.eventBroadcaster.emit('envelope', envelope);
            if (value_checker_1.doesHaveValue(envelope.testCase)) {
                this.remapDefinitionIds(envelope.testCase);
            }
            if (value_checker_1.doesHaveValue(envelope.testCaseFinished)) {
                this.parseTestCaseResult(envelope.testCaseFinished);
            }
        }
        else {
            throw new Error(`Unexpected message from worker: ${JSON.stringify(message)}`);
        }
    }
    saveDefinitionIdMapping(message) {
        lodash_1.default.each(message.stepDefinitionIds, (id, index) => {
            this.supportCodeIdMap[id] = this.supportCodeLibrary.stepDefinitions[index].id;
        });
        lodash_1.default.each(message.beforeTestCaseHookDefinitionIds, (id, index) => {
            this.supportCodeIdMap[id] = this.supportCodeLibrary.beforeTestCaseHookDefinitions[index].id;
        });
        lodash_1.default.each(message.afterTestCaseHookDefinitionIds, (id, index) => {
            this.supportCodeIdMap[id] = this.supportCodeLibrary.afterTestCaseHookDefinitions[index].id;
        });
    }
    remapDefinitionIds(testCase) {
        for (const testStep of testCase.testSteps) {
            if (testStep.hookId !== '') {
                testStep.hookId = this.supportCodeIdMap[testStep.hookId];
            }
            if (value_checker_1.doesHaveValue(testStep.stepDefinitionIds)) {
                testStep.stepDefinitionIds = testStep.stepDefinitionIds.map((id) => this.supportCodeIdMap[id]);
            }
        }
    }
    startWorker(id, total) {
        const workerProcess = child_process_1.fork(runWorkerPath, [], {
            cwd: this.cwd,
            env: lodash_1.default.assign({}, process.env, {
                CUCUMBER_PARALLEL: 'true',
                CUCUMBER_TOTAL_WORKERS: total,
                CUCUMBER_WORKER_ID: id,
            }),
            stdio: ['inherit', 'inherit', 'inherit', 'ipc'],
        });
        const worker = { closed: false, process: workerProcess };
        this.workers[id] = worker;
        worker.process.on('message', (message) => {
            this.parseWorkerMessage(worker, message);
        });
        worker.process.on('close', (exitCode) => {
            worker.closed = true;
            this.onWorkerProcessClose(exitCode);
        });
        const initializeCommand = {
            initialize: {
                filterStacktraces: this.options.filterStacktraces,
                supportCodePaths: this.supportCodePaths,
                supportCodeRequiredModules: this.supportCodeRequiredModules,
                options: this.options,
            },
        };
        worker.process.send(initializeCommand);
    }
    onWorkerProcessClose(exitCode) {
        if (exitCode !== 0) {
            this.success = false;
        }
        if (lodash_1.default.every(this.workers, 'closed')) {
            this.eventBroadcaster.emit('envelope', messages_1.messages.Envelope.fromObject({
                testRunFinished: {
                    timestamp: this.stopwatch.timestamp(),
                },
            }));
            this.onFinish(this.success);
        }
    }
    parseTestCaseResult(testCaseFinished) {
        const { worstTestStepResult } = this.eventDataCollector.getTestCaseAttempt(testCaseFinished.testCaseStartedId);
        if (!worstTestStepResult.willBeRetried &&
            this.shouldCauseFailure(worstTestStepResult.status)) {
            this.success = false;
        }
    }
    run(numberOfWorkers, done) {
        this.eventBroadcaster.emit('envelope', new messages_1.messages.Envelope({
            testRunStarted: {
                timestamp: this.stopwatch.timestamp(),
            },
        }));
        this.stopwatch.start();
        lodash_1.default.times(numberOfWorkers, (id) => this.startWorker(id.toString(), numberOfWorkers));
        this.onFinish = done;
    }
    giveWork(worker) {
        if (this.nextPickleIdIndex === this.pickleIds.length) {
            const finalizeCommand = { finalize: true };
            worker.process.send(finalizeCommand);
            return;
        }
        const pickleId = this.pickleIds[this.nextPickleIdIndex];
        this.nextPickleIdIndex += 1;
        const pickle = this.eventDataCollector.getPickle(pickleId);
        const gherkinDocument = this.eventDataCollector.getGherkinDocument(pickle.uri);
        const retries = helpers_1.retriesForPickle(pickle, this.options);
        const skip = this.options.dryRun || (this.options.failFast && !this.success);
        const runCommand = {
            run: {
                retries,
                skip,
                elapsed: this.stopwatch.duration().nanos(),
                pickle,
                gherkinDocument,
            },
        };
        worker.process.send(runCommand);
    }
    shouldCauseFailure(status) {
        return (lodash_1.default.includes([status_1.default.AMBIGUOUS, status_1.default.FAILED, status_1.default.UNDEFINED], status) ||
            (status === status_1.default.PENDING && this.options.strict));
    }
}
exports.default = Coordinator;
//# sourceMappingURL=coordinator.js.map